# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""robbie runtime environment module. This must be kept independent of robbie PySDK"""

from __future__ import absolute_import


# import logging
import sys
import shlex
import os
import subprocess
import threading
import select
import time
import yaml
import dataclasses
import json
import importlib.metadata
import platform
import argparse
import re
import json

# command runner modes
CONDA_MODE = "conda"
PYTHON_MODE = "python"
GENERIC_MODE = "generic"

if (os.getenv('POSITRON_CLOUD_ENVIRONMENT')):
    from positron_job_runner.cloud_logger import logger
    logger.debug("RuntimeEnvironmentManager: Running in cloud environment")
else:
    from common.logging_config import logger
    logger.debug("RuntimeEnvironmentManager: NOT running in cloud environment")


@dataclasses.dataclass
class _DependencySettings:
    """Dependency settings for the remote function.

    Instructs the runtime environment script on how to handle dependencies.
    If ``dependency_file`` is set, the runtime environment script will attempt
    to install the dependencies. If ``dependency_file`` is not set, the runtime
    environment script will assume no dependencies are required.
    """

    dependency_file: str = None

    def to_string(self):
        """Converts the dependency settings to a string."""
        return json.dumps(dataclasses.asdict(self))

    @staticmethod
    def from_string(dependency_settings_string):
        """Converts a json string to dependency settings.

        Args:
            dependency_settings_string (str): The json string to convert.
        """
        if dependency_settings_string is None:
            return None
        dependency_settings_dict = json.loads(dependency_settings_string)
        return _DependencySettings(dependency_settings_dict.get("dependency_file"))

    @staticmethod
    def from_dependency_file_path(dependency_file_path):
        """Converts a dependency file path to dependency settings.

        Args:
            dependency_file_path (str): The path to the dependency file.
        """
        if dependency_file_path is None:
            return _DependencySettings()
        if dependency_file_path == "auto-capture":
            return _DependencySettings("env_snapshot.yml")
        return _DependencySettings(os.path.basename(dependency_file_path))


class RuntimeEnvironmentManager:
    """Runtime Environment Manager class to manage runtime environment."""

    def snapshot(self, dependencies: str) -> str:
        """Creates snapshot of the user's environment

        If a req.txt or conda.yml file is provided, it verifies their existence and
        returns the local file path
        If ``auto_capture`` is set, this method will take the snapshot of
        user's dependencies installed in the local runtime.
        Current support for ``auto_capture``:
        * conda env, generate a yml file and return it's local path
        * pip env, generate a requirements.txt file and return it's local path

        Args:
            dependencies (str): dependencies file path or "auto-capture"

        Returns:
            file path of the existing or generated dependencies file
        """

        # No additional dependencies specified
        if dependencies is None:
            return None
        
        if dependencies == "auto-capture":
            if os.getenv("CONDA_DEFAULT_ENV"):
                # we are running in a conda environment
                return self._capture_from_conda_runtime()
            else:
                # we are not running on a conda environment
                return self._capture_from_pip_runtime()
        elif dependencies == "auto-capture-pip":
            return self._capture_deps_from_py_files()

        if dependencies == "none":
            return "none"
        # Dependencies specified as either req.txt or conda_env.yml
        if (
            dependencies.endswith(".txt")
            or dependencies.endswith(".yml")
            or dependencies.endswith(".yaml")
        ):
            self._is_file_exists(dependencies)
            return dependencies

        raise ValueError(f'Invalid dependencies provided: "{dependencies}"')

    def _pipdeptree_list(self, remove_robbie: bool = False) -> dict:
        
        def extract_packages(dependencies, package_dict):
            for dep in dependencies:
                package_dict[dep["package_name"]] = dep["installed_version"]
                extract_packages(dep["dependencies"], package_dict)

        # Get the dependency tree from pipdeptree
        result = subprocess.run(
                    "pipdeptree --json-tree",
                    capture_output=True,
                    text=True,
                    check=True,
                    shell=True,
                )
        dependency_tree = json.loads(result.stdout)

        # Extract remaining packages
        remaining_packages = {}
        extract_packages(dependency_tree, remaining_packages)

        # Define packages to remove (top-level only)
        if remove_robbie:
            if "robbie" in remaining_packages:
                del remaining_packages["robbie"]

        # For Windows local machines, remove pywin32 package
        if "pywin32" in remaining_packages:
            del remaining_packages["pywin32"]
        
        if "windows-curses" in remaining_packages:
            del remaining_packages["windows-curses"]
        
        return remaining_packages


    def _capture_from_pip_runtime(self, output_file = "requirements.txt") -> str:

        remaining_packages = self._pipdeptree_list(remove_robbie=True)

        # Write dependencies to requirements.txt
        with open(output_file, "w", encoding="utf-8") as file:
            for pkg, version in sorted(remaining_packages.items()):
                file.write(f"{pkg}=={version}\n")
       
        logger.debug(f"_capture_from_pip_runtime: returning")
        return os.path.join(os.getcwd(), output_file)

    def _capture_from_conda_runtime(self) -> str:
        """Generates dependencies list from the user's local runtime.

        Raises RuntimeEnvironmentError if not able to.

        Currently supports: conda environments
        """

        # Try to capture dependencies from the conda environment, if any.
        conda_env_name = self._get_active_conda_env_name()
        conda_env_prefix = self._get_active_conda_env_prefix()
        if conda_env_name:
            logger.info(f"Found conda_env_name: '{conda_env_name}'")
        elif conda_env_prefix:
            logger.info(f"Found conda_env_prefix: '{conda_env_prefix}'")
        else:
            raise ValueError("No conda environment seems to be active.")

        if conda_env_name == "base":
            logger.warning(
                "We recommend using an environment other than base to "
                "isolate your project dependencies from conda dependencies"
            )

        local_dependencies_path = os.path.join(os.getcwd(), "env_snapshot.yml")
        self._export_conda_env_from_prefix(conda_env_prefix, local_dependencies_path)

        return local_dependencies_path
    
    def _capture_deps_from_py_files(self) -> str:
        """
        Generates a requirement.txt from the .py files in the current directory
        See pipreqs package - https://github.com/bndr/pipreqs
        """
        logger.debug("Generating requirements.txt from the .py files in the current directory")
        _run_shell_cmd(f"pipreqs --force .")
        # this is a hack to remove the robbie package from the requirements.txt file
        _remove_package_from_requirements('robbie') 
        logger.debug(f"_capture_deps_from_py_files: returning")
        return os.path.join(os.getcwd(), "requirements.txt")

    def _get_active_conda_env_prefix(self) -> str:
        """Returns the conda prefix from the set environment variable. None otherwise."""
        return os.getenv("CONDA_PREFIX")

    def _get_active_conda_env_name(self) -> str:
        """Returns the conda environment name from the set environment variable. None otherwise."""
        return os.getenv("CONDA_DEFAULT_ENV")

    # @logger_module(module="Installing dependencies")
    def bootstrap(
        self, local_dependencies_file: str, client_python_version: str, conda_env: str = None
    ):
        """Bootstraps the runtime environment by installing the additional dependencies if any.

        Args:
            local_dependencies_file (str): path where dependencies file exists.
            conda_env (str): conda environment to be activated. Default is None.

        Returns: None
        """

        if local_dependencies_file.endswith(".txt"):
            if conda_env:
                self._install_req_txt_in_conda_env(conda_env, local_dependencies_file)
                #self._write_conda_env_to_file(conda_env)

            else:
                self._install_requirements_txt(local_dependencies_file, _python_executable())

        elif local_dependencies_file.endswith(".yml") or local_dependencies_file.endswith(".yaml"):
            """
            if conda_env:
                self._update_conda_env(conda_env, local_dependencies_file)
            else:
                conda_env = "robbie-runtime-env"
            """ 
            self._create_conda_env(conda_env, local_dependencies_file)
            # We will do this at a higher level depending upon the use case
            # self._validate_python_version(client_python_version, conda_env)
            # self._write_conda_env_to_file(conda_env)


    def change_dir_permission(self, dirs: list, new_permission: str):
        """Change the permission of given directories

        Args:
            dirs (list[str]): A list of directories for permission update.
            new_permission (str): The new permission for the given directories.
        """

        _ERROR_MSG_PREFIX = "Failed to change directory permissions due to: "
        command = ["sudo", "chmod", "-R", new_permission] + dirs
        logger.info(f"Executing {' '.join(command)}.")

        try:
            subprocess.run(command, check=True, stderr=subprocess.PIPE)
        except subprocess.CalledProcessError as called_process_err:
            err_msg = called_process_err.stderr.decode("utf-8")
            raise RuntimeEnvironmentError(f"{_ERROR_MSG_PREFIX} {err_msg}")
        except FileNotFoundError as file_not_found_err:
            if "[Errno 2] No such file or directory: 'sudo'" in str(file_not_found_err):
                raise RuntimeEnvironmentError(
                    f"{_ERROR_MSG_PREFIX} {file_not_found_err}. "
                    "Please contact the image owner to install 'sudo' in the job container "
                    "and provide sudo privilege to the container user."
                )
            raise RuntimeEnvironmentError(file_not_found_err)

    def _is_file_exists(self, dependencies):
        """Check whether the dependencies file exists at the given location.

        Raises error if not
        """
        if not os.path.isfile(dependencies):
            raise ValueError(f'No dependencies file named: {dependencies} was found.')

    def _install_requirements_txt(self, local_path, python_executable):
        """Install requirements.txt file"""
        #TODO: Add support for pipenv if necessary
        cmd = f"{python_executable} -m pip install -r {local_path} -U"
        logger.info(f"Running command: '{cmd}' in the dir: '{os.getcwd()}'")
        _run_shell_cmd(cmd)
        logger.info(f"Command {cmd} ran successfully")

    def _install_pip_package_from_dir(self, initial_cmd, local_path):
        """Install requirements.txt file"""
        #TODO: Add support for pipenv if necessary
        cmd = f"{initial_cmd} && {_python_executable()} -m pip install {local_path} -U"
        logger.info(f"Running command: {cmd} in the dir: {os.getcwd()}")
        _run_shell_cmd(cmd)
        logger.info(f"Command {cmd} ran successfully")


    def _create_conda_env(self, env_name, local_path):
        """Create conda env using conda yml file"""

        logger.info(f"_create_conda_env(): Current UID:{os.getuid()}, Current GID:{os.getgid()}")

        cmd = f"{self._get_conda_exe()} env create -n {env_name} --file {local_path}"
        # cmd = f"/opt/conda/bin/mamba env create -n {env_name} --file {local_path}"
        
        logger.info(f"Creating conda environment {env_name} using: {cmd}")
        _run_shell_cmd(cmd)
        logger.info(f"Conda environment: {env_name} created successfully.")

    def _install_req_txt_in_conda_env(self, env_name, local_path):
        """Install requirements.txt in the given conda environment"""

        cmd = f"{self._get_conda_exe()} run -n {env_name} pip install -r {local_path} -U"
        logger.info(f"Activating conda env and installing requirements: {cmd}")
        _run_shell_cmd(cmd)
        logger.info(f"Requirements installed successfully in conda env {env_name}")

    def _install_pip_package_from_dir_in_conda_env(self, env_name, package_dir):
        """Install pip package in the given conda environment"""

        cmd = f"{self._get_conda_exe()} run -n {env_name} pip install {package_dir}  -U"
        logger.info(f"Activating conda env and installing requirements: {cmd}")
        _run_shell_cmd(cmd)
        logger.info(f"Requirements installed successfully in conda env {env_name}")

    def _update_conda_env(self, env_name, local_path):
        """Update conda env using conda yml file"""

        cmd = f"{self._get_conda_exe()} env update -n {env_name} --file {local_path}"
        logger.info(f"Updating conda env: {cmd}")
        _run_shell_cmd(cmd)
        logger.info(f"Conda env {env_name} updated succesfully")

    '''
    def _export_conda_env_from_prefix(self, prefix, local_path):
        """Export the conda env to a conda yml file"""

        logger.info(f"Exporting conda environment: {prefix} to {local_path}")

        if platform.system() == "Windows":
            logger.info(f"Exporting Windows conda environment.")
            self._windows_conda_export(prefix, local_path)
        else:
            # cmd = f"{self._get_conda_exe()} env export -p {prefix} --no-builds > {local_path}"
            # logger.info(f"Exporting {platform.system()} conda environment: {cmd}")
            #_run_shell_cmd(cmd)
            self._nix_conda_export(prefix, local_path)
            logger.info(f"Conda environment {prefix} exported successfully")
    '''

    def _export_conda_env_from_prefix(self, prefix, output_file):
        """
        Export the current conda environment, including pip packages, to a specified YAML file.
        """
        # Construct the conda command based on the prefix (if provided)
        prefix_option = f"--prefix {prefix}"

        formatted_pip_packages = []
        pip_packages = self._pipdeptree_list(remove_robbie=False)
        for pkg, version in sorted(pip_packages.items()):
            formatted_pip_packages.append(f"      - {pkg.lower()}=={version}")

        # Export conda environment with history (without builds)
        history_env_proc = subprocess.run(
            "conda env export --from-history " + prefix_option,
            capture_output=True, text=True, check=True, shell=True
        )
        history_env = history_env_proc.stdout

        # Remove "prefix" line from history export
        history_env_lines = [line for line in history_env.splitlines() if not line.startswith("prefix:")]

        # Ensure the "- pip" line is present
        if not any(line.strip() == "- pip" for line in history_env_lines):
            history_env_lines.append("  - pip")
        history_env_lines.append("  - pip:")

        # Combine environment with pip packages
        environment_yml = "\n".join(history_env_lines) + "\n" + "\n".join(formatted_pip_packages)

        # Write to the specified file
        with open(output_file, "w", encoding="utf-8") as f:
            f.write(environment_yml)

        logger.info(f"Exported environment to: {output_file}")


    # TODO: remove this since its not used.
    def _windows_conda_export(self, prefix, output_file):
        """
        Export the current conda environment, including pip packages, to a specified YAML file.

        Args:
            prefix (str): The path to the conda environment. If None, uses the current active environment.
            output_file (str): The name of the output YAML file. Defaults to "environment.yml".
        """
        # Construct the conda command based on the prefix (if provided)
        prefix_option = f"--prefix {prefix}"

        # Export the full conda environment
        full_env_proc = subprocess.run(
            "conda env export " + prefix_option,
            capture_output=True, text=True, check=True, shell=True
        )
        full_env = full_env_proc.stdout

        # Extract pip packages
        pip_packages = []
        in_pip_section = False
        for line in full_env.splitlines():
            if "- pip:" in line:
                in_pip_section = True
                pip_packages.append(line)  # Keep the "- pip:" line
            elif in_pip_section:
                if line.startswith("      - "):  # Indented package names
                    if "pywin32" in line:  # Skip pywin32 package on Windows
                        continue
                    else:
                        pip_packages.append(line)
                else:
                    break  # End of pip section
        # Export conda environment with history (without builds)
        history_env_proc = subprocess.run(
            "conda env export --from-history " + prefix_option,
            capture_output=True, text=True, check=True, shell=True
        )
        history_env = history_env_proc.stdout

        # Remove "prefix" line from history export
        history_env_lines = [line for line in history_env.splitlines() if not line.startswith("prefix:")]

        # Ensure the "- pip" line is present
        if not any(line.strip() == "- pip" for line in history_env_lines):
            history_env_lines.append("  - pip")

        # Combine environment with pip packages
        environment_yml = "\n".join(history_env_lines) + "\n" + "\n".join(pip_packages)

        # Write to the specified file
        with open(output_file, "w", encoding="utf-8") as f:
            f.write(environment_yml)

        logger.info(f"Exported environment to: {output_file}")


    def _write_conda_env_to_file(self, env_name):
        """Writes conda env to the text file"""

        file_name = "robbie_conda_env.txt"
        file_path = os.path.join(os.getcwd(), file_name)
        with open(file_path, "w") as output_file:
            output_file.write(env_name)

    def _get_conda_exe(self):
        """Checks whether conda or mamba is available to use"""
        
        if (os.getenv('POSITRON_CLOUD_ENVIRONMENT')):
            logger.debug('_get_conda_exe(): running in cloud environment')
            return "/opt/conda/bin/mamba"
        else:
            # client 
            logger.debug("_get_conda_exe(): running in local or non 'job_user' cloud environment")
            import platform

            if platform.system() == "Windows":
                logger.debug("Running on Windows")
                which = "where"
            else:
                logger.debug("Not running on Windows")
                which = "which"

            if not subprocess.Popen([which, "mamba"], stdout=subprocess.PIPE, stderr=subprocess.PIPE).wait():
                return "mamba"
            if not subprocess.Popen([which, "conda"], stdout=subprocess.PIPE, stderr=subprocess.PIPE).wait():
                return "conda"
            raise ValueError("Neither conda nor mamba is installed or in the PATH")
            '''
            logger.debug'_get_conda_exe(): running in cloud environment')
            from .runner_env import runner_env 

            logger.debug"local environment: ", os.environ)
            logger.debug("runner_env: ", runner_env.env_without_runner_env())

            if (runner_env.JOB_USER == 'job_user'):
                logger.debug(f"_get_conda_exe(): Running as protected user")

                if not subprocess.Popen(
                    # ["which", "mamba"],
                    ["su", "-c", "test -f /opt/conda/bin/mamba", "-s", "/bin/bash", runner_env.JOB_USER],
                    env=runner_env.env_without_runner_env()
                ).wait():
                    return "mamba"
                raise ValueError("Neither conda nor mamba is installed on the image")


                result = subprocess.Popen(
                    ["su", "-c", "env", "-s", "/bin/bash", runner_env.JOB_USER], 
                    # capture_output=True, 
                    text=True,
                    env=runner_env.env_without_runner_env(),
                )
                if result.returncode == 0:
                    logger.debug(result.stdout)
                else:
                    logger.debug(f"print env ({result.returncode}): {result.stderr}")

                result = subprocess.Popen(
                    ["su", "-c", "which mamba", "-s", "/bin/bash", runner_env.JOB_USER], 
                    # capture_output=True, 
                    text=True,
                    env=runner_env.env_without_runner_env(),
                )
                if result.returncode == 0:
                    return result.stdout.strip()
                else:
                    logger.debug(f"Checking for mamba error ({result.returncode}): {result.stderr}")
        
                result = subprocess.Popen(
                    ["su", "-c", "which conda", "-s", "/bin/bash", runner_env.JOB_USER], 
                    # capture_output=True, 
                    text=True,
                    env=runner_env.env_without_runner_env(),
                )
                if result.returncode == 0:
                    return result.stdout.strip()
                else:
                    logger.debug(f"Checking for conda error ({result.returncode}): {result.stderr}")

                raise ValueError("Neither conda nor mamba is installed on the image")
                '''

    def _python_version_in_conda_env(self, env_name):
        """Returns python version inside a conda environment"""
        if not env_name:
            return None
        
        base_cmd = f"{self._get_conda_exe()} run -n {env_name} python --version"

        if (os.getenv('POSITRON_CLOUD_ENVIRONMENT')):
            logger.debug('_python_version_in_conda_env(): running in cloud environment')
            from .runner_env import runner_env 

            if (runner_env.JOB_USER == 'job_user'):
                logger.debug(f"_python_version_in_conda_env(): Running as protected user")
                cmd = f"su -c '{base_cmd}' -s /bin/bash job_user"
                try:
                    output = (
                        subprocess.check_output(shlex.split(cmd), 
                                                stderr=subprocess.STDOUT, 
                                                env=runner_env.env_without_runner_env(),
                                                cwd=runner_env.JOB_CWD)
                        .decode("utf-8")
                        .strip()
                    )
                    # convert 'Python 3.7.16' to [3, 7, 16]
                    version = output.split("Python ")[1].split(".")
                    return version[0] + "." + version[1]
                except subprocess.CalledProcessError as e:
                    raise RuntimeEnvironmentError(e.output)
                
        logger.debug(f"_python_version_in_conda_env(): Running as local or non-protected user")      
        try:
            output = (
                subprocess.check_output(base_cmd, stderr=subprocess.STDOUT, shell=True)
                .decode("utf-8")
                .strip()
            )
            # convert 'Python 3.7.16' to [3, 7, 16]
            version = output.split("Python ")[1].split(".")
            return version[0] + "." + version[1]
        except subprocess.CalledProcessError as e:
            raise RuntimeEnvironmentError(e.output)

    def _current_python_version(self):
        """Returns the current python version where program is running"""

        return f"{sys.version_info.major}.{sys.version_info.minor}".strip()

    def _current_robbie_pysdk_version(self):
        """Returns the current robbie python sdk version where program is running"""
        '''
        try:
            import robbie
            version = robbie.__version__
        except Exception as e:
            return f"Failed to retrieve robbie version due to: {e}"
        return version
        '''
        return importlib.metadata.version('robbie')

    def _validate_python_version(self, client_python_version: str, conda_env: str = None):
        """Validate the python version

        Validates if the python version where remote function runs
        matches the one used on client side.
        """
        if conda_env:
            job_python_version = self._python_version_in_conda_env(conda_env)
        else:
            job_python_version = self._current_python_version()
            
        if not job_python_version or client_python_version.strip() != job_python_version.strip():
            raise RuntimeEnvironmentError(
                f"Python version found in the container is '{job_python_version}' which "
                f"does not match python version '{client_python_version}' on the local client. "
                f"Please make sure that the python version used in the training container "
                f"is same as the local python version."
            )

    def _validate_robbie_pysdk_version(self, client_robbie_pysdk_version):
        """Validate the robbie python sdk version

        Validates if the robbie python sdk version where remote function runs
        matches the one used on client side.
        Otherwise, log a warning to call out that unexpected behaviors
        may occur in this case.
        """
        job_robbie_pysdk_version = self._current_robbie_pysdk_version()
        if (
            client_robbie_pysdk_version
            and client_robbie_pysdk_version != job_robbie_pysdk_version
        ):
            warning_str = f"""Inconsistent robbie versions found: 
                              robbie python sdk version found in the container is 
                              {job_robbie_pysdk_version}' which does not match the '{client_robbie_pysdk_version}' on the local client. 
                              Please make sure that the robbie version used in the training container 
                              is the same as the local robbie version in case of unexpected behaviors.
            """
            logger.warning(warning_str)                     
    
    def _list_conda_environments(self):
        cmd_list = [f'{self._get_conda_exe()}', 'env', 'list']
        try:
            # Run the 'conda env list' command and capture the output
            result = subprocess.run(cmd_list, stdout=subprocess.PIPE, text=True)
        
            # Extract lines from the output, skip header and empty lines
            lines = result.stdout.splitlines()
            envs = []
        
            # Parse the environments from the output
            for line in lines[2:]:  # Skip the header rows
                if line.strip():
                    env_name = line.split()[0]  # The environment name is the first element
                    envs.append(env_name)
        
            return envs
    
        except Exception as e:
            logger.debug(f"Error listing conda environments: {e}")
            return []
        
    def _valid_conda_env(self, conda_env: str) -> bool:
        if not conda_env:
            return False
        return conda_env in RuntimeEnvironmentManager()._list_conda_environments()

def _run_and_get_output_shell_cmd(cmd: str) -> str:
    """Run and return the output of the given shell command"""
    return subprocess.check_output(shlex.split(cmd), stderr=subprocess.STDOUT).decode("utf-8")


def _run_shell_cmd(cmd: str):
    """
    This method runs a given shell command using subprocess
    We handle the following cases:
    - Running in a local environment
    - Running in a cloud environment
        - As the positron user
        - As the protected user
    
    """
    # logger.debug(f"_run_shell_cmd(): {cmd}")
    # cwd is driven by the client or server side execution
    if os.getenv('POSITRON_CLOUD_ENVIRONMENT'):

        # logger.debug('_run_shell_cmd(): running in cloud environment')
        from .runner_env import runner_env 

        if (runner_env.JOB_USER != 'job_user'):
            # logger.debug('_run_shell_cmd(): as current user')
            p = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                shell=True,                # allows running on any os
                # universal_newlines=True,   # otherwise stdout and stderr will be bytes
                cwd=runner_env.JOB_CWD,
            )
        else:
            # logger.debug('_run_shell_cmd(): as protected user')

            env=runner_env.env_without_runner_env()

            # Unfortunately, the PATH variable is not  passed correctly to the subprocess in `env` with the su -c command
            # The workaround is to prefix the command with the PATH variable
            new_cmd = f"PATH={env['PATH']};{cmd}"
            # logger.debug(f"new_cmd: {new_cmd}")

            p = subprocess.Popen(
                ["su", "-c", new_cmd, "-s", "/bin/bash", "job_user"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                # requires elevated permissions which is not allowed in the runtime environment.
                # shell=True,
                cwd=runner_env.JOB_CWD,
                # universal_newlines=True,
                env=env,
            )
        log_stop_event = threading.Event()
        stdout_thread = threading.Thread(target=logging_thread, args=(p.stdout, "stdout", log_stop_event)); stdout_thread.start()
        stderr_thread = threading.Thread(target=logging_thread, args=(p.stderr, "stderr", log_stop_event)); stderr_thread.start()
        return_code = p.wait()
        log_stop_event.set()
        stdout_thread.join()
        stderr_thread.join()
        if return_code:
            error_message = f"Encountered error while running command '{cmd}'"
            raise RuntimeEnvironmentError(error_message)

    else:
        # logger.debug('_run_shell_cmd(): running in local environment')
        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            shell=True,                # allows running on any os
            # universal_newlines=True,   # otherwise stdout and stderr will be bytes
        )
        if platform.system() == 'Windows':
        # windows insanity
            try:
                # Read both stdout and stderr line by line
                for line in p.stdout:
                    print(line, end="")  # Avoid adding extra newlines
                for line in p.stderr:
                    print("ERROR: " + line, end="")
            except UnicodeEncodeError as e:
                print(f"Unicode error encountered: {e}")
            finally:
                p.stdout.close()
                p.stderr.close()
                return_code = p.wait()
            if return_code:
                error_message = f"Encountered error while running command '{cmd}'"
                raise RuntimeEnvironmentError(error_message)
        else:
            log_stop_event = threading.Event()
            stdout_thread = threading.Thread(target=logging_thread, args=(p.stdout, "stdout", log_stop_event)); stdout_thread.start()
            stderr_thread = threading.Thread(target=logging_thread, args=(p.stderr, "stderr", log_stop_event)); stderr_thread.start()
            return_code = p.wait()
            log_stop_event.set()
            stdout_thread.join()
            stderr_thread.join()
            if return_code:
                error_message = f"Encountered error while running command '{cmd}'"
                raise RuntimeEnvironmentError(error_message)

def logging_thread(pipe, stream_name: str, stop_event: threading.Event):
    try:
        with pipe:
            while not stop_event.is_set():
                # Calling pipe.readline() will hang the process until a new line is available
                # however, if no new lines are available, we want to check if the thread should stop
                ready, _, _ = select.select([pipe], [], [], 0.1)
                if ready:
                    # TODO: Can I iterate over the pipe and process everything that's ready?
                    line = pipe.readline()
                    if not line:
                        break
                    if line.strip():
                        if stream_name == "stdout":
                            logger.debug(line.rstrip())
                        else:
                            logger.debug("ERROR: " + line.rstrip())
                else:
                    stop_event.wait(1)

            # Read one last time.
            ready, _, _ = select.select([pipe], [], [], 0.1)
            if ready:
                for line in pipe:
                    if line.strip():
                        if stream_name == "stdout":
                            logger.debug(line.rstrip())
                        else:
                            logger.debug("ERROR: " + line.rstrip())
            # TODO: Do we need to check one more time for outstanding data in the pipe?
            # logger.debug(f'Logging thread for: {stream_name} stopped')
    except RuntimeEnvironmentError as e:
        logger.debug(f'Logging thread for: {stream_name} stopped with exception: {e}')

def _python_executable():
    """Return the real path for the Python executable, if it exists.

    Return RuntimeEnvironmentError otherwise.

    Returns:
        (str): The real path of the current Python executable.
    """
    if not sys.executable:
        raise RuntimeEnvironmentError(
            "Failed to retrieve the path for the Python executable binary"
        )
    return sys.executable


class RuntimeEnvironmentError(Exception):
    """The base exception class for bootstrap env excepitons"""

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

def _remove_package_from_requirements(package_name, file_path='./requirements.txt'):
    """ this is hack when the robbie package gets auto-added to the requirements.txt file in the auto-capture dependencies mode """
    # Read the file
    file_found = False
    try:
        with open(file_path, 'r') as file:
            lines = file.readlines()
            file_found = True
    except FileNotFoundError as e:
        return

    # Write back all lines except the one with the package name
    if not file_found:
        return
    try:
        with open(file_path, 'w') as file:
            for line in lines:
                # if not line.strip().startswith(package_name):
                if not package_name in line:
                    file.write(line)
    except Exception as e:
        return
    
import yaml

# TODO: remove, not used
def _remove_package_from_conda_yaml(file_path, package_name, output_path=None):
    """
    Removes a specific package from the dependencies of a Conda YAML file.
    
    Args:
        file_path (str): The path to the Conda YAML file.
        package_name (str): The name of the package to remove.
        output_path (str, optional): The path to save the modified YAML file. 
                                     If None, it will overwrite the original file.
    """
    # Step 1: Load the YAML file
    with open(file_path, 'r') as file:
        conda_yaml = yaml.safe_load(file)
    
    # Step 2: Modify the dependencies
    dependencies = conda_yaml.get('dependencies', [])
    
    # Handle dependencies that may be listed as dictionaries (for pip) or strings
    new_dependencies = []
    for dep in dependencies:
        if isinstance(dep, str):  # e.g., 'numpy=1.21.0'
            if not dep.startswith(package_name + '=') and dep != package_name:
                new_dependencies.append(dep)
        elif isinstance(dep, dict) and 'pip' in dep:  # pip dependencies are listed as a dictionary
            pip_dependencies = dep.get('pip', [])
            filtered_pip_dependencies = [pip_dep for pip_dep in pip_dependencies if not pip_dep.startswith(package_name + '=') and pip_dep != package_name]
            if filtered_pip_dependencies:
                new_dependencies.append({'pip': filtered_pip_dependencies})
        else:
            new_dependencies.append(dep)
    
    # Step 3: Update the conda YAML content
    conda_yaml['dependencies'] = new_dependencies

    # Step 4: Write the updated YAML file
    if output_path is None:
        output_path = file_path  # Overwrite the original file if no output path is provided
    
    with open(output_path, 'w') as file:
        yaml.dump(conda_yaml, file)
    
    logger.debug(f"Package '{package_name}' removed successfully from {output_path}.")

    
def _build_job_args(
    rpv: str, 
    dependencies: str,
    mode: str = None,
    debug: str = None,
):
    """
    Builds the job arguments for the container runtime

    Args:
        mode (str): The type of command runner to use
            - CONDA_MODE: Conda job (Python or non-Python) with dependencies defined by .yml
            - PYTHON_MODE: Non-conda Python job with dependencies defined by .txt or auto-captured
            - GENERIC: Shells commands to run without regard for Python or Conda
        rpv (str): The Python version to use in the container
        dependencies (str): The path to the dependencies file
    """
    job_args = []
    if not mode == None: 
        if mode not in [CONDA_MODE, PYTHON_MODE, GENERIC_MODE]:
            raise ValueError(f"Invalid command runner type: {mode}")
        job_args = ["--mode", mode]

    job_args.extend(
        [
        "--client_python_version",
        RuntimeEnvironmentManager()._current_python_version(),
        ]
    )

    job_args.extend(
        [
        "--rpv",
        rpv,
        ]
    )

    job_args.extend(
        [
        "--client_robbie_pysdk_version",
        RuntimeEnvironmentManager()._current_robbie_pysdk_version(),
        ]
    )

    job_args.extend(
        [
         "--dependency_settings",
            _DependencySettings.from_dependency_file_path(
                dependencies
            ).to_string(),
        ]
    )
    # if running under a conda env, send that over to the job
    if os.getenv("CONDA_DEFAULT_ENV"):
        job_args.extend(["--local_conda_env", os.getenv("CONDA_DEFAULT_ENV")])


    return job_args

def _running_in_conda():
    return True if os.getenv("CONDA_DEFAULT_ENV") else False


def _is_python_package_in_conda_yaml(file_path: str) -> bool:
    try:
        # Load the YAML file
        with open(file_path, 'r') as file:
            conda_env = yaml.safe_load(file)

        # Check if 'dependencies' key is in the file
        dependencies = conda_env.get('dependencies', [])
        
        # Search for the python package in dependencies
        for dep in dependencies:
            # Dependencies can be either strings or dictionaries for pip packages
            if isinstance(dep, str) and dep.startswith("python"):
                return True
            elif isinstance(dep, dict) and "python" in dep.get("pip", []):
                return True

        # If no python package found
        return False

    except Exception as e:
        logger.debug(f"Error reading or parsing {file_path}: {e}")
        return False
    
    import yaml
from typing import Optional

def _get_python_version_from_conda_yaml(file_path: str) -> Optional[str]:
    try:
        # Load the YAML file
        with open(file_path, 'r') as file:
            conda_env = yaml.safe_load(file)

        # Check if 'dependencies' key is in the file
        dependencies = conda_env.get('dependencies', [])

        # Search for the python package in dependencies
        for dep in dependencies:
            # Dependencies can be strings or dictionaries for pip packages
            if isinstance(dep, str) and dep.startswith("python"):
                # Extract version if specified (e.g., "python=3.8")
                if "=" in dep:
                    return dep.split("=")[1]  # Returns the version part
                else:
                    return "Version not specified"
            elif isinstance(dep, dict) and "pip" in dep:
                # Check pip dependencies
                for pip_dep in dep["pip"]:
                    if pip_dep.startswith("python"):
                        if "==" in pip_dep:
                            return pip_dep.split("==")[1]
                        else:
                            return "Version not specified"

        # Return None if no Python package is found
        return None

    except Exception as e:
        logger.debug(f"Error reading or parsing {file_path}: {e}")
        return None
    
        









import os
from common.console import console, ROBBIE_BLUE, ROBBIE_DEFAULT
from rich.prompt import Confirm
from prompt_toolkit import prompt
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import Validator
from prompt_toolkit.completion import WordCompleter
from common.config import PositronJob, PositronJobConfig
from common.enums import JobRunType
from common.utils import _current_python_version, _valid_python_versions
from common.image_name import get_auto_image_name_and_cluster
from positron_job_runner.runtime_environment_manager import (
     RuntimeEnvironmentManager,
    _running_in_conda,
    _get_python_version_from_conda_yaml,
    CONDA_MODE,
    PYTHON_MODE,
    GENERIC_MODE
)
from cli.helpers import (
    FundingSources,
    Environments,
    Images
)

def prompt_and_build_positron_job_config(
    job_config_yaml: PositronJob = None,
    cmd_runner_dash_i: bool = False
) -> PositronJobConfig:
    """
    Prompts the user, in the command line, for parameters to run a job and returns a PositronJobConfig object.
    This function is invoked in two ways:
    - from `run.py` - for command line or notebook runner jobs - `cmd_runner_dash_i` is True
    - from `configure.py` - to build job_config.yaml for all job types (Runner and Remote Function) - `cmd_runner_dash_i` is False
    """
 
    while True:
        try:

            pj = PositronJob()

            style = Style.from_dict({
                'completion-menu.completion': 'bg:#008888 #ffffff',
                'completion-menu.completion.current': 'bg:#00aaaa #000000',
            })

            if not cmd_runner_dash_i:
                # User selects the job type
                job_type_list = [
                    "Command Runner",
                    "Remote Function"
                ]
                def is_job_type_valid(text):
                    return text == "" or text in job_type_list

                job_type_validator = Validator.from_callable(
                    is_job_type_valid, 
                    error_message="Please choose a valid job type."
                )
                job_type_selection = prompt(
                    message=_blue_msg("Choose a job type", f"{job_type_list[0]}"),
                    completer=WordCompleter(job_type_list), 
                    style=style, 
                    validator=job_type_validator
                )
                if job_type_selection == "":
                    pj.job_type = JobRunType.BASH_COMMAND_RUNNER
                elif job_type_selection == job_type_list[0]:
                    pj.job_type = JobRunType.BASH_COMMAND_RUNNER
                elif job_type_selection == job_type_list[1]:
                    pj.job_type = JobRunType.REMOTE_FUNCTION_CALL
                else:
                    console.print("[yellow]Invalid job type, aborting and starting over.[/yellow]")
                    continue
            else:
                # we are in command line mode, so we know the job type
                pj.job_type = JobRunType.BASH_COMMAND_RUNNER


            # now ask for the mode depdending on the job type
            if pj.job_type == JobRunType.BASH_COMMAND_RUNNER:
                mode_list = [
                    "Generic",
                    "Conda (Python and Non-Python)", 
                    "Python (Non-Conda)", 
                ]
 
                def is_mode_valid(text):
                    return text == "" or text in mode_list

                mode_validator = Validator.from_callable(
                    is_mode_valid, 
                    error_message="Please choose a valid mode."
                )
                mode_selection = prompt(
                    message=_blue_msg(
                        "Choose an environment for the remote machine:", 
                        f"{mode_list[0]}"
                    ),
                    completer=WordCompleter(mode_list), 
                    style=style, 
                    validator=mode_validator
                )
                if mode_selection == "":
                    pj.mode = GENERIC_MODE
                elif mode_selection == mode_list[0]:
                    pj.mode = GENERIC_MODE
                elif mode_selection == mode_list[1]:
                    pj.mode = CONDA_MODE
                elif mode_selection == mode_list[2]:
                    pj.mode = PYTHON_MODE
                else:
                    console.print("[red]Invalid mode, aborting and starting over.[/red]")
                    continue

            elif pj.job_type == JobRunType.REMOTE_FUNCTION_CALL:
                # remote function
                mode_list = [
                    "Conda (Python and Non-Python)",
                    "Python (Non-Conda)", 
                ]
                def is_mode_valid(text):
                    return text == "" or text in mode_list

                mode_validator = Validator.from_callable(
                    is_mode_valid, 
                    error_message="Please choose a valid mode."
                )
                mode_selection = prompt(
                    message=_blue_msg(
                        "Choose an environment for the remote machine:", 
                        f"{mode_list[0]}"
                    ),
                    completer=WordCompleter(mode_list), 
                    style=style, 
                    validator=mode_validator
                )
                if mode_selection == "":
                    pj.mode = CONDA_MODE
                elif mode_selection == mode_list[0]:
                    pj.mode = CONDA_MODE
                elif mode_selection == mode_list[1]:
                    pj.mode = PYTHON_MODE
                else:
                    console.print("[red]Invalid mode, aborting and starting over.[/red]")
                    continue
            else:
                console.print("[red]Invalid job type, aborting and starting over.[/red]")
                continue

            #
            # Conda
            #
            if pj.mode == CONDA_MODE:
                # Let the user select the conda environment, either the current environment or from a yaml file

                # if _running_in_conda():
                #    console.print(f"Conda environment detected: {os.getenv('CONDA_DEFAULT_ENV')}") 

                conda_env_list = _get_files_with_extensions(".", [".yml", ".yaml"])
                if _running_in_conda():
                    conda_env_list.extend([f"auto-capture"])
        
                def is_conda_env_valid(text):
                    return text == "" or text in conda_env_list

                conda_env_validator = Validator.from_callable(
                    is_conda_env_valid, 
                    error_message='Please choose a valid conda environment or file.'
                )

                conda_env = prompt(
                    message=_blue_msg("Choose a conda environment [.yml/.yaml] file", f"auto-capture"),
                    completer=WordCompleter(conda_env_list), 
                    style=style, 
                    validator=conda_env_validator
                )
 
                # Only indicate we have deps if the user selects something
                if len(conda_env):
                    pj.dependencies = conda_env
                    pj.dep_selection = "user selected"
                elif conda_env == "":
                    pj.dependencies = "auto-capture"
                    pj.dep_selection = "default"
                else:
                    console.print(f"[red]Error: You must select a valid conda environment or file.")
                    continue

            elif pj.mode == PYTHON_MODE:

                if pj.job_type == JobRunType.BASH_COMMAND_RUNNER:
                # Only allow Python version changes in the command runner mode        
                    def is_python_version_valid(text):
                        return text == "" or text == "local" or text in _valid_python_versions()

                    python_version_validator = Validator.from_callable(
                        is_python_version_valid, 
                        error_message="Please choose a valid Python version or none."
                    )
                    wclist = _valid_python_versions()
                    wclist.append("local")

                    python_version_selection = prompt(
                        message=_blue_msg("Choose Python version", 
                                          "local"
                        ),
                        completer=WordCompleter(wclist), 
                        style=style, 
                        validator=python_version_validator
                    )
                    # If the user explicitly selects a different version, set it in the job config
                    if (python_version_selection == "" or 
                        python_version_selection == "local"
                    ):
                        pj.rpv = "local"
                    else:
                        pj.rpv = python_version_selection
                        pj.rpv_selection = "user selected"
                elif pj.job_type == JobRunType.REMOTE_FUNCTION_CALL:
                    pj.rpv = "local"
                else:
                    console.print("[red]Invalid job type, aborting and starting over.")
                    continue
                # 
                # Dependencies
                # 
                # python deps - requirements.txt or auto-capture
                py_deps_list = _get_files_with_extensions(".", [".txt"])
                if (pj.rpv == _current_python_version() or 
                    pj.rpv == "local"):
                    py_deps_list.extend(["auto-capture","none"])
                else:
                    py_deps_list.extend(["none"])
        
                def is_dependency_valid(text):
                    return text == "" or text in py_deps_list

                dependency_validator = Validator.from_callable(
                    is_dependency_valid, 
                    error_message='Please choose a valid selection.'
                )

                deps = prompt(
                    message=_blue_msg(
                        "Choose Python dependencies: ", 
                        "requirements.txt"
                    ),
                    completer=WordCompleter(py_deps_list), 
                    style=style, 
                    validator=dependency_validator
                )
                # only indicate we have deps if the user selects something
                if len(deps):
                    pj.dependencies = deps
                    pj.dep_selection = "user selected"
                elif deps == "":
                    pj.dependencies = "requirements.txt"
                    pj.dep_selection = "default"
                else:
                    console.print(f"[red]Error: You must select a valid Python dependency file.")
                    continue
            elif pj.mode == GENERIC_MODE:
                # default out for image selection
                pj.rpv = "local"
            else:
                console.print("[yellow]Invalid mode, aborting and starting over.[/yellow]")
                continue

            # Get the user's funding sources
            try:
                fs = FundingSources.load()
                if fs == None:
                    console.print(f"[bold red]User has no funding sources, please contact: support@robbie.run")
                    return None
            except Exception as e:
                console.print(f"[bold red]Unable to fetch funding sources: {e}") 
                return None

            # validate the user input
            def is_fs_valid(text):
                return text == "" or text in fs.menu_items()
            
            fs_validator = Validator.from_callable(
                is_fs_valid, 
                error_message='Please select a valid funding source.'
            )

            fs_choice = prompt(
                message=_blue_msg("Select how to bill your job", "Personal tokens"),
                completer=WordCompleter(fs.menu_items()), 
                style=style, 
                validator=fs_validator
            )

            if len(fs_choice):
                fs_id = fs.id_from_menu_item(fs_choice)
                pj.funding_selection = "user selected"
            else:
                fs_id = fs.default_funding_source_id()
                pj.funding_selection = "default"

            pj.funding_group_id = fs_id

            #
            # Environments
            #

            # are there any environments in this funding source?
            try:
                envs = Environments.load(pj.funding_group_id)
                if envs == None:
                     # no environments for the user oh well
                    console.print(f"[red]Error your environments: {fs_choice} has no approved hardware, please contact 'support@robbie.run'")
                    return None
            except Exception as e:
                console.print(f"[bold red]Unable to fetch environments: {e}") 
                return None

            # validate the user input
            def is_env_valid(text):
                return text == "" or text in envs.menu_items() or text == "Automatically select using AI (experimental)"
            env_validator = Validator.from_callable(
                is_env_valid, 
                error_message='Please enter a valid environment.'
            )

            env_menu_items = envs.menu_items()
            env_menu_items.append("Automatically select using AI (experimental)")

            env_choice = prompt(
                message=_blue_msg("Select your preferred hardware", 
                                    ("None" 
                                    if not fs.default_env_name_from_fs_id(fs_id) 
                                    else fs.default_env_name_from_fs_id(fs_id))
                ), 
                completer=WordCompleter(env_menu_items), 
                style=style, 
                validator=env_validator)

            if len(env_choice):
                if env_choice == "Automatically select using AI (experimental)":
                    pj.environment_id = "auto-selected"
                    pj.environment_selection = "auto-selected (experimental)"
                else:
                    pj.environment_id = envs.id_from_menu_item(env_choice)
                    pj.environment_selection = "user selected"
            else:
                # choose the default, if available
                if fs.default_env_id_from_fs_id(fs_id) == None:
                    console.print(f"[red] Error funding source: {fs_choice} has no default hardware and you didn't specify any.")
                    return None
                else:
                    pj.environment_id = fs.default_env_id_from_fs_id(fs_id)
                    pj.environment_selection = "default"

            # 
            # Include Local directory 
            #
            pj.include_local_dir = False
            text = prompt(
                message=_blue_msg(f"Copy local CWD to be copied to the remote machine?", "[y]/n"),
                style=style)
            if len(text):
                if text.lower() == "y":
                    pj.include_local_dir = True
            else:
                pj.include_local_dir = True
        
            # advanced options include: image selecdtion, max tokens, max duration, and environment variables
            if Confirm.ask("Configure advanced options?", default=False):
                if pj.include_local_dir:
                    #
                    # Custom file filters
                    #
                    first_pass = True
                    while True:
                        ignore_patten = prompt('Enter file patterns to ignore (Enter a <blank> line to go to the next step):',
                                          style=Style.from_dict({'prompt': ROBBIE_BLUE})
                        )
                        if ignore_patten == "":
                            break
                        if first_pass:
                            pj.custom_file_filter = []
                            first_pass = False
                        pj.custom_file_filter.append(ignore_patten)

                #
                # Images
                #
                images = Images.load(pj.funding_group_id, pj.environment_id)
                def is_image_valid(text):
                    return (text == "" 
                            or text == "auto-select" 
                            or text in images.menu_items()
                    )

                image_validator = Validator.from_callable(
                    is_image_valid, 
                    error_message='Please choose a valid image.'
                )

                if images:
                    image_list = images.menu_items()
                    image_list.append("auto-select")
                    image_choice = prompt(message=_blue_msg("Select your preferred image", "auto-select"),
                                          completer=WordCompleter(image_list), 
                                          style=style, 
                                          validator=image_validator
                    )

                    if image_choice == "" or image_choice == "auto-select":
                        # the actual image selection occurs in `run.py`
                        pj.image = "auto-select"
                        pj.image_selection = "default"
                    elif len(image_choice) and image_choice != "auto-select":
                        # the user hit tab and selected an image
                        pj.image = images.name_from_menu_item(image_choice)
                        # we just get the cluster type from the environment
                        pj.cluster = envs.cluster_type_from_env_id(pj.environment_id)
                        pj.image_selection = "user selected"
                else:
                    # no environments for the user oh well
                    console.print(f"[red]Failed to get images, please contact 'support@robbie.run'")
                    return None
            
                # 
                # Max tokens
                #
                def is_max_token_valid(text):
                    return text == "" or (text.isdigit() and int(text) >=1 and int(text) < 10000)

                max_token_validator = Validator.from_callable(is_max_token_valid, 
                                                              error_message='Please enter a number between 1 and 10000.'
                )

                text = prompt(
                    message=_blue_msg("Maximum tokens to consume", "none"),
                    style=style, 
                    validator=max_token_validator
                )
                if len(text):
                    pj.max_tokens = int(text)

                # 
                # Max duration
                #
                def is_max_duration_valid(text):
                    return text == "" or (text.isdigit() and int(text) >=1 and int(text) < 1000)
                duration_validator = Validator.from_callable(is_max_duration_valid, error_message='Please enter a valid duration between 1 and 1000.')

                text = prompt(
                    message=_blue_msg("Max duration in minutes", "none"),
                    style=style, 
                    validator=duration_validator)

                if len(text):
                    pj.max_time = int(text)

                # 
                # Environment variables
                #
                preserve = False
                if job_config_yaml and job_config_yaml.env:
                    console.print(f"[yellow]{job_config_yaml.filename} contains the following environment variables:")
                    for env in job_config_yaml.env:
                        console.print(f"[yellow]     {env}")
                    response = prompt('Reuse for this run? [[y]/n]:', style=Style.from_dict({'prompt': 'yellow'}))
                    if response in ["y", "yes", "Yes", "Y", ""]:
                        preserve = True  
                        pj.env = job_config_yaml.env 

                if not preserve:
                    first_pass = True
                    while True:
                        var_name = prompt('Environment variable name (Enter a <blank> line to go to the next step):',
                                          style=Style.from_dict({'prompt': ROBBIE_BLUE})
                        )
                        if not var_name:
                            break
                        var_value = prompt(f'Value for {var_name} (hint= Enter a <blank> line to use local machine value):', style=style)
                        if first_pass:
                            pj.env = {}
                            first_pass = False
                        pj.env[var_name] = var_value
            else:
                pj.image = "auto-select"
                pj.image_selection = 'Automatically Selected'

            # 
            # Commands - loop through and capture the commands
            #
            if pj.job_type == JobRunType.BASH_COMMAND_RUNNER:
                msg = "Enter the commands to run on the remote machine (Enter a <blank> line when you are done entering commands):"
                preserve = False
                if job_config_yaml and job_config_yaml.commands:
                    console.print(f"[yellow]{job_config_yaml.filename} contains the following commands:")
                    for cmd in job_config_yaml.commands:
                        console.print(f"[yellow]     {cmd}")
                    response = prompt('Reuse for this run? [[y]/n]:', style=Style.from_dict({'prompt': 'yellow'}))
                    if response in ["y", "yes", "Yes", "Y", ""]:
                        preserve = True
                        pj.commands = job_config_yaml.commands        

                if not preserve:
                    first_pass = True
                    while True:
                        cmd = prompt(msg, style=Style.from_dict({'prompt': ROBBIE_BLUE}))
                        if not cmd:
                            break
                        if first_pass:
                            pj.commands = []
                            first_pass = False
                        pj.commands.append(cmd)

            # user can enter a name
            name_choice = prompt(
                _blue_msg("Please enter a custom run name", "Robbie Generated"),
                style=style)
            if len(name_choice):
                pj.name = name_choice

            # print(f"End of Interactive Job Config: {pj.to_string()}")

            return PositronJobConfig(version="1.1", python_job=pj)

        except (KeyboardInterrupt, EOFError) as e:
            console.print("[yellow]Aborting job configuration.")
            return None


def _blue_msg(message, default):
    """
    Helper function to build a prompt with the ROBBIE blue color
    """
    msg = [ 
        (ROBBIE_BLUE, f"{message} ["),
        (ROBBIE_DEFAULT, default),
        (ROBBIE_BLUE, "]: " )
        ]
    return msg


def _get_files_with_extensions(directory, extensions_to_include):
    """Gets a list of files in the directory, include  specified extensions."""
    all_files = os.listdir(directory)
    # use list comprehension to filter out files with the wrong extensions
    filtered_files = [
        file
        for file in all_files
        if os.path.splitext(file)[1].lower() in extensions_to_include
    ]
    return filtered_files


def get_job_config_yaml_name():

    style = Style.from_dict({
        'completion-menu.completion': 'bg:#008888 #ffffff',
        'completion-menu.completion.current': 'bg:#00aaaa #000000',
    })

    def is_job_config_valid(text):
        return text == "" or text == "e" or text.endswith(".yaml") or text.endswith(".yml")
    
    jc_validator = Validator.from_callable(
        is_job_config_valid, 
        error_message='Please enter a valid file name!'
    )

    while True:
        jc_name = prompt(
            message=_blue_msg("Specify a file name to save your job configuration or (e) to exit without saving: ", "job_config.yaml"), 
            style=style, 
            validator=jc_validator
        )
        if jc_name == "e":
            return None
        if jc_name == "":
            jc_name = "job_config.yaml"
    
        if os.path.isfile(jc_name):
            if Confirm.ask(f"File {jc_name} already exists. Overwrite?", default=False):
                return jc_name
            else:
                console.print("[yellow]Try again.")
        else:
            return jc_name




    




import os
import signal
import json
import time
import sentry_sdk
import secrets
import threading
from rich.prompt import Confirm
from remote_function.stored_function import StoredFunction
from common.validation import validate_python_version
from common.utils import _exit_by_mode, SUCCESS, FAILURE, _nb
from common.exceptions import RemoteFunctionException, RemoteCallException
from common.custom_file_filter import (
    CustomFileFilter,
    copy_workdir
)

import common.api.stream_logs
from common.config import PositronJob
from common.enums import JobStatus, failed_job_statuses
from common.api.stream_logs import start_stdout_stream
from common.api.get_job import get_job_status, get_job
from common.api.start_job import start_job
from common.api.terminate_job import terminate_job
from common.api.create_job import create_job
from common.cli_args import args as cli_args
from common.job import _prepare_and_upload_workspace
from common.env_config import env
from common.user_config import user_config
from common.console import console, ROBBIE_BLUE
from common.logging_config import logger
from common.api.terminate_job import terminate_job
from common.print_banner import (
    print_robbie_configuration_banner, 
    print_job_details_banner, 
    print_known_error, 
    print_job_complete_banner
)

from common.local_container import run_local_docker_container

# TODO: update env var to just be base url
PORTAL_BASE = env.API_BASE.rstrip('/api')

# number of seconds to poll when monitoring job status
POLLING_SEC=1

class Deploy:
    @staticmethod
    def remote_function_deploy(func, args, kwargs, job_config: PositronJob):
        global job
        # signal.signal(signal.SIGINT, handle_sigint)
        try:
            # TODO: lots of this code is the same as the other `deploy` file, need to consolidate
            logger.debug(f'Job Config: {job_config}')
            job_config.validate_values()
            logger.debug(f'Runtime Environment: {job_config.create_runtime_env()}')

            if not validate_python_version(job_config.image):
                _exit_by_mode(FAILURE)
                return

            job = create_job(job_config)

            if job_config.custom_file_filter:
                custom_file_filter = CustomFileFilter(ignore_name_patterns=job_config.custom_file_filter)
            else:
                custom_file_filter = None

            # this just builds the tree so we can print it out
             # actual files are copied later
            copy_workdir(".", custom_file_filter, print_tree=True)

            print_robbie_configuration_banner(job, job_config)
            logger.debug(f"Created Run Details: {json.dumps(job, indent=2)}")

            # prompt the user if they don't pass the -y option
            if not cli_args.skip_prompts:
                user_input = input("Run with these settings? (Y/n)")
                if not user_input.lower() in ["", "yes", "y", "Yes", "Y"]:
                    terminate_job(job["id"], "User declined from CLI")
                    console.print("[yellow]See you soon![/yellow]")
                    return

            console.print("\n[bold]Local Machine: [/bold]", style=ROBBIE_BLUE)    

            # create a random key for the function integrity check
            hmac_key = secrets.token_hex(32)
            # hack to get the secret key into the environment when running a local container
            os.environ['REMOTE_FUNCTION_SECRET_KEY'] = hmac_key

            console.print("[green]✔[/green] [1/4] Serialized function.", style=ROBBIE_BLUE)
            # Create stored function from func and arguments
            stored_function = StoredFunction(func, args, kwargs)
            stored_function.set_job_id(job['id'])
            stored_function.serialize_function()
            logger.debug(f"### create_function_metadata(): user_config.user_auth_token = {user_config.user_auth_token}")
            stored_function.create_function_metadata(hmac_key=hmac_key)
            stored_function.upload_to_s3()

            console.print("[green]✔[/green] [2/4] Packaged up workspace artifacts.", style=ROBBIE_BLUE)
        
            # now upload workspace (current directory and deps)
            _prepare_and_upload_workspace(
                job_id = job['id'], 
                local_deps_path = job_config.dependencies,
                include_local_workdir = True if job_config.include_local_dir else False,
                pre_execution_commands = None,
                pre_execution_script_local_path = None,
                # custom_file_filter = None,
                custom_file_filter = custom_file_filter
            )
            console.print("[green]✔[/green] [3/4] Uploaded compressed workspace to Robbie.", style=ROBBIE_BLUE)

            if cli_args.create_only:
                console.print(f"[green]✔[/green] [4/4] Run created successfully.")
                console.print(f"JOB_ID: {job.get('id')}")
                if cli_args.local_container:
                    run_local_docker_container(job.get('id'), cli_args.local_container)
                _exit_by_mode(SUCCESS)
                return

            # set the secret key
            runtime_env = job_config.create_runtime_env()
            runtime_env['REMOTE_FUNCTION_SECRET_KEY'] = hmac_key
            
            logger.debug(f"Runtime Environment: {runtime_env}")
             # start job
            start_job(job_id=job['id'], data=runtime_env)
            console.print("[green]✔[/green] [4/4] Submitted run to Robbie.", style=ROBBIE_BLUE)
            
            start = time.perf_counter()
            print_job_details_banner(job)
            console.print(f"You can also monitor run status in the Robbie portal at: {PORTAL_BASE}/portal/app/my-runs?jobId={job['id']}\n", style=ROBBIE_BLUE)

            # Start standard output stream if option selected
            console.print("Waiting for remote job to finish...", style=ROBBIE_BLUE)

            if cli_args.stream_stdout:
                thread = threading.Thread(target=print_job_status, args=(job,))
                # job status printing thread
                thread.start()
                # start the stdout stream
                common.api.stream_logs.start_stdout_stream(job['id'], job_config.verbose)
                thread.join()

            while True:
                job = get_job_status(job['id'])
                if job['status'] == JobStatus.complete:
                    console.print(f"[green]✔[/green] Done! Now processing results...")
                    result, exception = stored_function.load_and_validate_results(hmac_key=hmac_key)
                    console.print("Passing back result.")
                    if exception is not None:
                        raise RemoteFunctionException(exception)
                    else:
                        return result
                elif job['status'] == JobStatus.failed or job['status'] == JobStatus.execution_error:
                    raise RemoteCallException(user_friendly_message="The remote run has failed, please check job output logs for further details!")
                elif job['status'] == JobStatus.terminated:
                    # users terminated
                    break
                elif job['status'] == JobStatus.terminate:
                    # we terminated
                    break
                time.sleep(2)

            # get the job info a final time
            job = get_job(job['id'])
            print_job_complete_banner(job, start)
            # print details if error for debugging.
            if job['status'] in failed_job_statuses:
                console.print(f"We are sorry that your job has run in to an issue. If you continue to have issues, please contact support@robbie.run and provide the following traceback.\nTrace ID: {sentry_sdk.get_current_span().trace_id}")
            _exit_by_mode(SUCCESS)

        except KeyboardInterrupt:
            if Confirm.ask("Interrupt received, do you want to terminate the run?", default=False):
                console.print("[yellow]Terminating run...[/yellow]")
                terminate_job(job['id'], "User interrupted")
            
        except RemoteCallException as e:
            """For known errors we dont print exceptions, we just print the user friendly message"""
            print_known_error(e)
            _exit_by_mode(FAILURE)
        except RemoteFunctionException as e:
            """
                When the result of a remote function call is an Exception being raised
                this is how we let the deserialized exception bubble up back to the client side
            """
            raise e.rf_exception
        except Exception:
            console.print_exception()


def _is_job_done(job) -> bool:
    return (
        job['status'] == "terminated" or 
        job['status'] == "complete" or 
        job['status'] == "failed" or 
        job['status'] == "execution_error" or
        job['status'] == "terminate_job"
    )

def print_job_status(job):
    """
    Print the job status to the console
    """
    last_status_change = "Starting..."
    start_time = time.time()

    console.print("Processing...", style=ROBBIE_BLUE)   

    try:
        while True:
            job = get_job_status(job['id'])
            # are we in a final state?
            # print("job status: ", job['status'], "is_set: ", common.api.stream_logs.stop_flag.is_set())
            if(_is_job_done(job) or common.api.stream_logs.stop_flag.is_set()):
                break

            # there has been a status change
            if(job['status'] != last_status_change):
                # take care of the previous status
                if last_status_change != "Starting...":
                    elapsed_time = time.time() - start_time
                    minutes = int(elapsed_time // 60)
                    seconds = int(elapsed_time % 60)
                    tenth_of_seconds = int((elapsed_time * 10) % 10)
                    if minutes > 0:
                        timer_text = "{:3}m {:2}.{:1}s".format(minutes, seconds, tenth_of_seconds)
                    else:
                        timer_text = "{:2}.{:1}s".format(seconds, tenth_of_seconds)
                    console.print(timer_text)

                # print out the new job status
                if job['status'] == "started_container":
                    console.print(f" => booting_runtime...", end="")
                else:
                    console.print(f" => {job['status']}...", end="")
                last_status_change = job['status']
                start_time = time.time()
            time.sleep(POLLING_SEC)

    except Exception as e:
        console.print(f"print_job_status error: {e}")  
    finally:
        logger.debug("Done monitoring status...")



import time
import signal
import threading
from rich.prompt import Confirm
from common.exceptions import RemoteCallException
from common.custom_file_filter import (
    CustomFileFilter,
    copy_workdir
)
from common.env_config import env
from common.config import PositronJob
from common.cli_args import args as cli_args
from common.print_banner import *
from common.console import console, ROBBIE_BLUE
from cli.cmds.download import download_chosen_results
from common.job import _prepare_and_upload_workspace
from common.logging_config import logger
import common.api.stream_logs
from common.api.get_job import get_job, get_job_status
from common.api.start_job import start_job
from common.api.terminate_job import terminate_job
from common.api.create_job import create_job
from common.utils import _exit_by_mode, SUCCESS, FAILURE
from common.validation import warn_python_version
from common.local_container import run_local_docker_container

# number of seconds to poll when monitoring job status
POLLING_SEC=1

# for the the deep link
PORTAL_BASE = env.API_BASE.rstrip('/api')
    
def command_runner_deploy(job_config: PositronJob):
    """
    This function is the entry point for the running a command on a remote machine.
    It can be called in two ways:
    - Mode 1 - From the command line with the `positron run` command
    - Mode 2 - From the run_notebook() function when running in a notebook (Jupyter, Colab, VScode
    """
    
    try:
        logger.debug(job_config.to_string("command_runner_deploy"))
        job_config.validate_values()
        
        warn_python_version(job_config.image)

        # TODO: We should not be creating a job before we let the user run it, we need defaults in the DBs that we can query
        logger.debug(job_config.create_runtime_env())
        job = create_job(job_config=job_config)

        # we want to print out the files that will be copied over
        if job_config.custom_file_filter:
            custom_file_filter = CustomFileFilter(ignore_name_patterns=job_config.custom_file_filter)
        else:
            custom_file_filter = None
        
        # this just builds the tree so we can print it out
        # actual files are copied later
        copy_workdir(".", custom_file_filter, print_tree=True)

        print_robbie_configuration_banner(job, job_config)

        # prompt the user if they don't pass the -y option
        if not cli_args.skip_prompts:
            if not Confirm.ask("Run with these settings?", default=True):
                terminate_job(job["id"], "User declined from CLI")
                console.print("[yellow]See you soon![/yellow]")
                return

        # tell people we are on the local machine
        console.print("[bold]Local Machine: [/bold]", style=ROBBIE_BLUE)    

        console.print("[green]✔[/green] [1/3] Packaged up workspace artifacts.", style=ROBBIE_BLUE)

        _prepare_and_upload_workspace(
            job_id = job['id'], 
            local_deps_path = job_config.dependencies,
            include_local_workdir = True if job_config.include_local_dir else False,
            pre_execution_commands = None,
            pre_execution_script_local_path = None,
            custom_file_filter = custom_file_filter,
        )
        console.print("[green]✔[/green] [2/3] Uploaded compressed workspace to Robbie.", style=ROBBIE_BLUE)
        
        if cli_args.create_only:
            console.print(f"[green]✔[/green] [3/3] Job created successfully", style=ROBBIE_BLUE)
            console.print(f"JOB_ID: {job.get('id')}")
            if cli_args.local_container:
                run_local_docker_container(job.get('id'), cli_args.local_container)
            return

        # start the job up
        start_job(job_id=job['id'], data=job_config.create_runtime_env())
        console.print("[green]✔[/green] [3/3] Submitted job to Robbie.", style=ROBBIE_BLUE)

        start = time.perf_counter()
        print_job_details_banner(job)
        console.print(f"You can also monitor job status in the Robbie portal at: {PORTAL_BASE}/portal/app/my-runs?jobId={job['id']}\n", style=ROBBIE_BLUE) 

        # Are we streaming stdout or just showing the status changes.
        if cli_args.stream_stdout:
            # tell people we are on the remote machine
            console.print("[bold]Remote Machine Status: [/bold]", style=ROBBIE_BLUE)  

            thread = threading.Thread(target=print_job_status, args=(job,))
            thread.start()

            common.api.stream_logs.start_stdout_stream(job['id'], job_config.verbose)
            
            thread.join()

            final_get_job = get_job(job['id'])
            # did someone interrupt the stream?
            if _is_job_done(final_get_job):
                print_job_complete_banner(final_get_job, start)
                if cli_args.download:
                    # download the results
                    download_chosen_results(final_get_job['id'], cli_args.download, cli_args.local_path)
            
                if _was_job_a_success(final_get_job):
                    _exit_by_mode(SUCCESS)
                else:
                    _exit_by_mode(FAILURE)
            else:
                console.print(f"You can monitor job status in the Robbie portal at: {PORTAL_BASE}/portal/app/my-runs?jobId={job['id']}", style=ROBBIE_BLUE) 
                _exit_by_mode(SUCCESS)
    
    except KeyboardInterrupt:
        if Confirm.ask("Interrupt received, do you want to terminate the run?", default=False):
            console.print("[yellow]Terminating run...[/yellow]")
            if job is not None:
                terminate_job(job['id'], "User interrupted")
            # the CLI will exit when the main loop sees the status change.
        else:
            console.print("[yellow]Exiting...run will continue. Please monitor in the portal.[/yellow]")
            _exit_by_mode(SUCCESS)
    except RemoteCallException as e:
        """For known errors we dont print exceptions, we just print the user friendly message"""
        print_known_error(e)
        _exit_by_mode(FAILURE)
    except Exception as e:
        # don't let this propagate up, we want to catch all exceptions
        logger.exception(e)
        print(e)
        _exit_by_mode(FAILURE)

def _is_job_done(job) -> bool:
    return (
        job['status'] == "terminated" or 
        job['status'] == "complete" or 
        job['status'] == "failed" or 
        job['status'] == "execution_error" or
        job['status'] == "terminate_job"
    )

def _was_job_a_success(job) -> bool:
    return job['status'] == "complete"

def print_job_status(job):
    """
    Print the job status to the console
    """
    last_status_change = "Starting..."
    start_time = time.time()

    console.print("Processing...", style=ROBBIE_BLUE)   

    try:
        while True:
            job = get_job_status(job['id'])
            # are we in a final state?
            # print("job status: ", job['status'], "is_set: ", common.api.stream_logs.stop_flag.is_set())
            if(_is_job_done(job) or common.api.stream_logs.stop_flag.is_set()):
                break

            # there has been a status change
            if(job['status'] != last_status_change):
                # take care of the previous status
                if last_status_change != "Starting...":
                    elapsed_time = time.time() - start_time
                    minutes = int(elapsed_time // 60)
                    seconds = int(elapsed_time % 60)
                    tenth_of_seconds = int((elapsed_time * 10) % 10)
                    if minutes > 0:
                        timer_text = "{:3}m {:2}.{:1}s".format(minutes, seconds, tenth_of_seconds)
                    else:
                        timer_text = "{:2}.{:1}s".format(seconds, tenth_of_seconds)
                    console.print(timer_text)

                # print out the new job status
                # print out the new job status
                if job['status'] == "started_container":
                    console.print(f" => booting_runtime...", end="")
                else:
                    console.print(f" => {job['status']}...", end="")
                last_status_change = job['status']
                start_time = time.time()
            time.sleep(POLLING_SEC)

    except Exception as e:
        console.print(f"print_job_status error: {e}")  
    finally:
        logger.debug("Done monitoring status...")


import requests
import json
from common.exceptions import RemoteCallException
from ..env_config import env
from ..logging_config import logger

def get_upload_presigned_url(job_id, filename):
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN, "PositronJobId": job_id}
    url = f'{env.API_BASE}/generate-presigned-url?filename={filename}'
    logger.debug(f'Calling: {url}')
    response = requests.get(url, headers=Headers)
    logger.debug(response)
    if response.status_code != 200:
        raise RemoteCallException(f'Presigned url fetching failed with http code: {response.status_code} \n {response.text}')
    else:
        logger.debug(response.json())
        return response.json()
    
def get_download_presigned_url(job_id, filename):
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN, "PositronJobId": job_id}
    url = f'{env.API_BASE}/generate-download-url?filename={filename}'
    logger.debug(f'Calling: {url}')
    response = requests.get(url, headers=Headers)
    logger.debug(response)
    if response.status_code != 200:
        raise RemoteCallException(
            f'Presigned url fetching failed with http code: {response.status_code} \n {response.text}')
    else:
        logger.debug(response.json())
        return response.json()


import requests
import json
from common.exceptions import RemoteCallException
from ..env_config import env
from ..logging_config import logger

def get_job(job_id):
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN, "PositronJobId": job_id}
    url = f'{env.API_BASE}/get-job'
    
    logger.debug(f'Calling: {url}')
    response = requests.get(url, headers=Headers)
    
    logger.debug(response)
    if response.status_code != 200:
        body = response.json()
        logger.debug(json.dumps(body, indent=2))
        if body.get('userFriendlyErrorMessage'):
            raise RemoteCallException(body.get('userFriendlyErrorMessage'))
        else:
            raise RemoteCallException(f'get job failed with http code: {response.status_code} \n {response.text}')
    else:
        logger.debug(json.dumps(response.json(), indent=2))
        return response.json()


def get_job_status(job_id):
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN, "PositronJobId": job_id}
    url = f'{env.API_BASE}/get-job-status'

    logger.debug(f'Calling: {url}')
    response = requests.get(url, headers=Headers)

    logger.debug(response)
    if response.status_code != 200:
        body = response.json()
        logger.debug(json.dumps(body, indent=2))
        if body.get('userFriendlyErrorMessage'):
            raise RemoteCallException(body.get('userFriendlyErrorMessage'))
        else:
            raise RemoteCallException(f'get job status failed with http code: {response.status_code} \n {response.text}')
    else:
        logger.debug(json.dumps(response.json(), indent=2))
        return response.json()


import os
import typer
import pyfiglet
import re
import pprint
import platform
import tempfile
import shutil
from git import Repo
from rich.prompt import Confirm
from rich.spinner import Spinner
from rich.live import Live
from rich.text import Text
from typing_extensions import Annotated
from common.console import console
from common.constants import JOB_CONF_YAML_PATH
from common.observability.main import track_command_usage
from cli.interactive import (
    prompt_and_build_positron_job_config,
    get_job_config_yaml_name
)
from common.console import console, ROBBIE_BLUE, ROBBIE_DEFAULT
from common.api.get_env_rec import get_env_rec
from common.api.funding_envs_images import *
from cli.auto_complete import (
    funding_group_auto_complete,
    job_config_auto_complete,
    cluster_auto_complete,
)
from common.config import (
    load_job_config, 
    PositronJob,
    PositronJobConfig
)
from common.logging_config import logger


SPINNER = "line" if platform.system() == "Windows" else "dots"

@track_command_usage("rec")
def rec(
    python_file: Annotated[str, typer.Argument(help='Python file to analyze')],
    job_config_arg: Annotated[str, 
                            typer.Option("--f", 
                                        help="Specify the job configuration file.",
                                        autocompletion=job_config_auto_complete)] = "./job_config.yaml",
    funding_arg: Annotated[str, typer.Option("--funding_group_id", 
                                help="Specify the team to run the job on.",
                                autocompletion=funding_group_auto_complete)] = None,
    cluster_arg: Annotated[str, 
                        typer.Option("--cluster",
                                        help="Specify the cluster to run the job on.", 
                                        autocompletion=cluster_auto_complete)] = None

) -> None:
    """
    Use GenAI to get an environment recommendation
    """

    ascii_banner = pyfiglet.figlet_format("Robbie")
    console.print(ascii_banner, style='#41a7ff')

    console.print(f"Robbie will use AI to determine the best hardware for you Python program.\n", style=ROBBIE_BLUE)

    global_job_config_yaml = None

    try:
        if funding_arg is None:
        # try to determine FG from job config
            global_job_config_yaml = load_job_config(job_config_arg)
            if global_job_config_yaml:
                logger.debug(f"global_job_config_yaml: {global_job_config_yaml.to_string()}")
                console.print(f"[green]✔[/green] Determining funding_group_id from configuration file: ", end="", style=ROBBIE_BLUE)
                console.print(f"{job_config_arg}", style=ROBBIE_DEFAULT)
                if global_job_config_yaml.funding_group_id:
                    funding_group_id = global_job_config_yaml.funding_group_id
                else:
                    # we need to get the default funding group
                    funding_group_id = _get_personal_fg()
                    if funding_group_id is None:
                        console.print(f"[red]No funding group found. Please specify a funding group.")
                        return
                    console.print(f"[green]✔[/green] Using personal funding group: {funding_group_id}", style=ROBBIE_BLUE)
                    
            else:
                logger.debug(f"{job_config_arg} not found.")
                # we need to get the default funding group
                funding_group_id = _get_personal_fg()
                if funding_group_id is None:
                    console.print(f"[red]No funding group found. Please specify a funding group.")
                    return
                console.print(f"[green]✔[/green] Using personal funding group: {funding_group_id}", style=ROBBIE_BLUE)
                    
        else:
            funding_group_id = funding_arg
            console.print(f"[green]✔[/green] Using funding group argument: {funding_group_id}", style=ROBBIE_BLUE)
                    

        system_prompt, human_prompt = build_prompt_from_envs(funding_group_id, python_file, cluster_arg)
        if human_prompt == "":
            console.print(f"[red]Error - no files to analyze.")
            return

    except Exception as e:
        print(f'Error: {str(e)}')
        return

    with Live(Spinner(SPINNER, text=Text(f"Securely analyzing your code...", style=ROBBIE_BLUE)), refresh_per_second = 20, console=console, transient=True):
        try:
            return_env = get_env_rec(
                system_prompt=system_prompt,
                human_prompt=human_prompt
            )
        except Exception as e:
            console.print(f"Error getting environment recommendation: {str(e)}")
            return

    console.print(f"[green]✔[/green] Analysis Complete!", style=ROBBIE_BLUE)

    console.print(f"[green]✔[/green] Recommanded Environment: ", style=ROBBIE_BLUE)

    console.print(return_env.get("response"))

    uuid = find_uuids(return_env.get("response"))
    if len(uuid)== 1:
        env_id = find_uuids(return_env.get("response"))[0]
    else:
        console.print(f"Error: Could not find a correct number of UUIDs in the response.")
        return

    if Confirm.ask("\nDo you want to save this environment in your job config?", default=False):
        if global_job_config_yaml is None:
            console.print(f"[red]Error: couldn't add environment to existing job configuration file: {job_config_arg}")
            return
        filename = get_job_config_yaml_name()
        if filename:
            global_job_config_yaml.environment_id = env_id
            pjc = PositronJobConfig(version="1.1", python_job=global_job_config_yaml)
            pjc.write_to_file(filename=filename)
            console.print(f"[green]✔[/green] Success!", style=ROBBIE_BLUE)
    return

def extract_gpu_details(gpu_string: str):
    # Enhanced regex to handle more variations in GPU naming
    match = re.search(r'(?P<gpu_type>.+)-(?P<vram>\d+GB)$', gpu_string)
    if match:
        gpu_type = match.group('gpu_type')
        vram = match.group('vram')
        return gpu_type, vram
    else:
        return None, None

# Test with multiple GPU strings
gpu_list = [
    "NVIDIA-A100-SXM4-40GB",
    "NVIDIA-V100-PCIe-16GB",
    "NVIDIA-RTX3090-24GB",
    "NVIDIA-H100-PCIe-80GB",
    "NVIDIA-Tesla-K80-12GB"
]


def get_aws_gpu_details(instance_type: str):
    """
    Returns the GPU type and VRAM size (as a string) for a given AWS instance type.

    Args:
        instance_type (str): The AWS instance type (e.g., 'p4d.24xlarge').

    Returns:
        tuple: (GPU type, VRAM size as a string) or ('Unknown GPU', '0GB') if not found.
    """
    # Mapping of AWS instance families to GPU type and VRAM size (as string)
    aws_gpu_mapping = {
        "p4d": ("NVIDIA A100", "40GB"),
        "p4de": ("NVIDIA A100", "40GB"),
        "p3": ("NVIDIA V100", "16GB"),
        "p3dn": ("NVIDIA V100", "32GB"),
        "p2": ("NVIDIA K80", "12GB"),
        "g4dn": ("NVIDIA T4", "16GB"),
        "g5": ("NVIDIA A10G", "24GB"),
        "g5g": ("NVIDIA T4G", "16GB"),
        "g3": ("NVIDIA M60", "8GB"),
        "g2": ("NVIDIA K520", "4GB"),
        "p5": ("NVIDIA H100", "80GB"),
        "g6": ("NVIDIA L4 Tensor Core", "24GB"),
        "p6": ("NVIDIA H100", "80GB"),
        "g4ad": ("AMD Radeon Pro V520", "8GB"),  # Not NVIDIA, included for completeness
        "dl1": ("Habana Gaudi", "32GB"),        # Not NVIDIA
        "trn1": ("AWS Trainium", "0GB"),        # Not NVIDIA
        "inf1": ("AWS Inferentia", "0GB"),      # Not NVIDIA
    }

    # Extract the portion of the instance name before the dot
    instance_prefix = instance_type.split('.')[0]

    # Retrieve GPU details or return defaults if not found
    return aws_gpu_mapping.get(instance_prefix, ("Unknown GPU", "0GB"))


def _get_personal_fg():
    """
    Get the personal funding group
    """
    fs = list_funding_sources()
    if len(fs) == 0:
        return None
    
    personal_fs_id = None
    for key, val in fs.items():
        if (val[FS_TYPE] == FS_PERSONAL_TYPE):
            personal_fs_id = val.get(FS_ID)
    
    if personal_fs_id is None:
        return None
    else:
        return personal_fs_id


def find_uuids(text):
    """
    Finds all UUIDs in a string.

    Args:
        text: The string to search.

    Returns:
        A list of UUIDs found in the string.
    """
    uuid_pattern = r"[0-9a-f]{8}-[0-9a-f]{4}-[0-5][0-9a-f]{3}-[089ab][0-9a-f]{3}-[0-9a-f]{12}"
    return re.findall(uuid_pattern, text, re.IGNORECASE)


def build_prompt_from_envs(funding_group_id: str, python_file: str, cluster_arg: str = None):

    if not is_git_url(python_file):
        if python_file != ".":
            if not os.path.exists(python_file):
                console.print(f"[red]Python file not found: {python_file}")
                raise FileNotFoundError(f"Python file not found: {python_file}")

            if not python_file.endswith(".py") and not python_file.endswith(".ipynb"):
                console.print(f"[red]Python or notebook file must end with .py or .ipynb: {python_file}")
                raise Exception("Python file must end with .py or .ipynb")

    python_code = None

    try:
        if (is_git_url(python_file)):
            python_code = fetch_python_files_from_repo(python_file)
            if python_code == "":
                console.print(f"[red]Error fetching Python files from repo: {python_file}")
                raise Exception(f"Error fetching Python files from repo: {python_file}")
        elif python_file == ".":  # read all python files in the current directory
            python_code = read_python_files()
        elif python_file.endswith(".ipynb") or python_file.endswith(".py"):
            console.print(f"  Files for analysis:", style=ROBBIE_BLUE)
            console.print(f"[yellow] - {python_file}")
            with open(python_file, 'r') as file:
                python_code = file.read()
        else:
            console.print(f"[red]Python file not found: {python_file}")
            raise FileNotFoundError(f"Python file not found: {python_file}")

        # lets build the list of environments to for the prompt
        envs = list_environments(fs_id=funding_group_id)
        if envs is None:
            console.print(f"No environments found for this funding group: {funding_group_id}")
            raise Exception(f"No environments found for this funding group: {funding_group_id}")

        env_entry_list = []
        for key, val in envs.items():
            # print(f"Environment: {val}")
            if cluster_arg is not None and val.get(ENV_CLUSTER_TYPE) != cluster_arg:
                continue
            if not val.get(ENV_DELETED):
                if (val.get(ENV_GPU_NUMBER) == 0 or val.get(ENV_GPU_NUMBER) == None or val.get(ENV_GPU_NUMBER) == ""):
                    gpu_number = "None"
                    gpu_type = "None"
                    gpu_vram = "None"
                else:
                    gpu_number = val.get(ENV_GPU_NUMBER)
                    if val.get(ENV_CLUSTER_TYPE) == "NERC":
                        gpu_type, gpu_vram = extract_gpu_details(val.get(ENV_GPU_TYPE))
                        if gpu_type is None or gpu_vram is None:
                            logger.warning(f"Could not extract GPU details from: {val.get(ENV_GPU_TYPE)}")
                            gpu_type = "None"
                            gpu_vram = "None"
                    elif val.get(ENV_CLUSTER_TYPE) == "EKS":
                        instance_type = val.get(ENV_NODE_TYPE)
                        if instance_type is None or instance_type == "":
                            console.print(f"Instance type not found for environment: {val.get(ENV_ID)}")
                            raise Exception(f"Instance type not found for environment: {val.get(ENV_ID)}")
                        gpu_type, gpu_vram = get_aws_gpu_details(instance_type)
                        if gpu_type is None or gpu_vram is None:
                            logger.warning(f"Could not extract GPU details from: {val.get(ENV_GPU_TYPE)}")
                            gpu_type = "None"
                            gpu_vram = "None"
                    elif val.get(ENV_CLUSTER_TYPE) == "DOCKER":
                        continue
                    else:
                        console.print(f"Unknown cluster type: {val.get(ENV_CLUSTER_TYPE)}")
                        raise Exception(f"Unknown cluster type: {val.get(ENV_CLUSTER_TYPE)}")

                # print(f"id: {val.get(ENV_ID)}, GPU Type: {gpu_type}, VRAM: {gpu_vram}, GPU Number: {gpu_number}\n\n")
                entry = { 
                    "id": val.get(ENV_ID), 
                    "environmentName": val.get(ENV_NAME), 
                    "cpu": val.get(ENV_CPU),
                    "ram": val.get(ENV_RAM),
                    "disk": val.get(ENV_DISK),
                    "gpuNumber": gpu_number,
                    "gpuType": gpu_type,
                    "gpuVRAM": gpu_vram,
                    "tokensPerHour": val.get(ENV_TPH),
                }
                env_entry_list.append(entry)
        if len(env_entry_list) == 0:
            console.print(f"No environments found for this funding group: {funding_arg}")
            raise Exception(f"Instance type not found for environment: {val.get(ENV_ID)}")
    except Exception as e:
        console.print(f'Error loading environments! {str(e)}')
        raise e
        

    # print(pprint.pformat(env_entry_list))

    system_prompt = f"""
You are an automated code analysis tool that looks a Python machine learning source code and makes a recommendation of the hardware to use for training.
Your will choose from a list of available hardware options. 
Each entry in the list will have the following JSON structure.

    "id": a UUID of the hardware, 
    "environmentName": a string representing the name of the hardware,
    "cpu": a string representing the number of virtual CPUs,
    "ram": a string representing the amount of system ram in gigabyes represented as Gi,
    "disk": a string representing the amount of disk space in gigabyes represented as Gi,
    "gpuNumber": a string representing the number of GPUs, "None" if no GPU, 
    "gpuType": a string representing the type of GPU if one exists,
    "gpuVRAM": a string representing the amount of VRAM on the GPU in GB,
    "tokensPerHour": the cost per hour in token to use the hardware

Here is the list: {str(env_entry_list)}
Your goal is to choose the hardware that will train the model the fastest while keeping the cost as low as possible.
I am going to provide Python source code and I want you to return the name, cost, GPU type, GPU VRAM, and UUID of the environment. After a carriage return provide a three sentence explanation of you chose the hardware option.
"""

    human_prompt = f"""{python_code}"""

    return system_prompt, human_prompt


def fetch_python_files_from_repo(repo_url):
    """
    Fetches all python files from a github repo, clones it to a temporary directory,
    reads all of the .py files and returns a string with the files contents.
    
    args:
    repo_url (str): The URL of the GitHub repository.
    
    returns:
    str: A string containing the combined content of all .py files in the repository,
         or an empty string if an error occurs.
    """
    console.print(f"  Files to analyze: {repo_url}", style=ROBBIE_BLUE)
    temp_dir = tempfile.mkdtemp()
    try:
        Repo.clone_from(repo_url, temp_dir)
        python_files_content = ""
        for root, _, files in os.walk(temp_dir):
            for file in files:
                if file.endswith(".py") or file.endswith(".ipynb"):
                    console.print(f"[yellow]  - {file}")
                    file_path = os.path.join(root, file)
                    with open(file_path, "r") as f:
                        python_files_content += f.read() + "\n"
        return python_files_content
    except Exception as e:
         print(f"An error occurred: {e}")
         return ""
    finally:
        shutil.rmtree(temp_dir)


def read_python_files(directory="."):
    """
    Recursively searches for .py files in the given directory and returns their contents as a string.

    Args:
        directory (str, optional): The directory to search in. Defaults to the current directory.

    Returns:
        str: A string containing the contents of all .py files found, separated by newlines.
    """
    console.print(f"  Files to analyze in: {os.getcwd()}", style=ROBBIE_BLUE)
    all_contents = ""
    for root, _, files in os.walk(directory):
        for file in files:
            if file.endswith(".py") or file.endswith(".ipynb"):
                console.print(f"[yellow]  - {file}")
                filepath = os.path.join(root, file)
                try:
                    with open(filepath, "r") as f:
                        contents = f.read()
                        all_contents += contents + "\n"
                except Exception as e:
                    print(f"Error reading file {filepath}: {e}")
    return all_contents


def is_git_url(url: str) -> bool:
    """
    Detects if a string is a valid URL and if it ends with '.git'.

    Args:
        url (str): The URL string to check.

    Returns:
        bool: True if the string is a URL and ends with '.git', False otherwise.
    """
    # Regex pattern to check for valid URL ending with '.git'
    pattern = r'^(https?|git|ssh|ftp)://[^\s]+\.git$'
    return bool(re.match(pattern, url))




import typer
from cli.helpers import (
    FundingSources,
    Environments,
    Images,
)
from cli.interactive import _get_files_with_extensions
from common.utils import is_valid_uuid
from positron_job_runner.runtime_environment_manager import (
    _running_in_conda,
    RuntimeEnvironmentManager
)

from rich.console import Console
err_console = Console(stderr=True)   
 # err_console.print(f"ctx.params={ctx.params}") 

"""
Auto complete functions for the `robbie run` command

These implement Typer specific autocompletion logic

https://typer.tiangolo.com/tutorial/options-autocompletion/#access-other-cli-parameters-with-the-context

"""
def job_config_auto_complete(ctx: typer.Context):
    """
    Job config for the CLI autocomplete
    """
    jc_files = _get_files_with_extensions(".", [".yaml"])
    return jc_files

def cluster_auto_complete(ctx: typer.Context):
    """
    cluster CLI autocomplete
    """
    return ["NERC", "EKS"]
    
def funding_group_auto_complete(ctx: typer.Context):
    """
    Funding groups for the CLI autocomplete
    """
    # check if `--environment_id` was already specified
    env = ctx.params.get('environment_arg')
    if (env):
        err_console.print(f"Error: please specify --funding-group first")
        return []
    try:
        fs = FundingSources.load()
        if fs == None:
            return []
        else:
            return fs.auto_complete_items()
    except Exception as e:
        err_console.print(f"[bold red]Auto complete error: {e}") 
        return [" "]
    
def environment_auto_complete(ctx: typer.Context):
    """
    Environments for the CLI autocomplete
    """
    fga = ctx.params.get('funding_arg')
    if (fga):
        # --funding_arg was previously specified
        try:
            envs = Environments.load(fga)
            if envs == None:
                err_console.print(f"Error: There are no environments in group: {fga}.")
                return []
            else:
                return envs.auto_complete_items()
        except Exception as e:
            err_console.print(f"[bold red]Auto complete error: {e}") 
            return [" "]
    else:
        # get the default FG
        try:
            fs = FundingSources.load()
            if fs == None:
                err_console.print(f"Error: Your are not a member of a Group")
                return []
            else:
                envs = Environments.load(fs.default_funding_source_id())
                if envs == None:
                    err_console.print(f"Error: Your have no enviroments in your Personal group.")
                    return []
                else:
                    return envs.auto_complete_items()
        except Exception as e:
            err_console.print(f"[bold red]Auto complete error: {e}") 
        return [" "]
    
    
def images_auto_complete(ctx: typer.Context):
    """
    Images for the CLI autocomplete
    """
    env = ctx.params.get('environment_arg')
    fg = ctx.params.get('funding_arg')
    if fg and env:
        try:
            images = Images.load(fg, env)
            if images == None:
                err_console.print(f"Error: No images are available.")
                return []
            else:
                img_list = images.auto_complete_items()
                img_list.append("auto-select")
                return img_list
        except Exception as e:
            err_console.print(f"[bold red]Auto complete error: {e}") 
        return [" "]
    
def deps_auto_complete(ctx: typer.Context):
    """
    Dependencies autocomplete
    """
    ca = ctx.params.get('conda_arg')
    if ca:
        conda_env_list = _get_files_with_extensions(".", [".yml", ".yaml"])
        if _running_in_conda():
            conda_env_list.extend([f"auto-capture"])
        return conda_env_list

    pv = ctx.params.get('python_ver_arg')
    pa = ctx.params.get('python_arg')
    if pv and not pa:
        err_console.print(f"[bold red]Can't specify version, specifying --python first")
        return [" "] 
    if pa:
        py_deps_list = _get_files_with_extensions(".", [".txt"])
        if pv:
            if pv == RuntimeEnvironmentManager()._current_python_version():
                py_deps_list.extend(
                        ["auto-capture", 
                        "none"]
            )
            else:
                py_deps_list.extend(["none"])
        else:
            py_deps_list.extend(
                ["auto-capture", 
                "none"]
            )
        return py_deps_list

    return [" "]
    
def set_env_auto_complete():
    return ["local", "dev", "alpha", "beta"]


def jobs_auto_complete(incomplete: str):
    from cli.cmds.download import Jobs
    try:
        jobs = Jobs.load()
        if (jobs):
            completion = []
            # returns list of tuples (name, id)
            for name in jobs.auto_complete_items():
                if name[0].startswith(incomplete):
                    completion.append(name)
            return completion
    except Exception as e:
        err_console.print(f"[bold red]Auto complete error: {e}") 
        return [" "]

    